/*
Icmpinfo
by Luigi Auriemma
e-mail: aluigi@autistici.org
web:    aluigi.org

This tool lets everyone to see the current time and possibly the
netmask of a remote host.

The time routine manages both the 2 "stack" methods used by the
remote hosts (for example Win shows time in a different byte order
than Linux).

The function sndrcv_icmp has been written to add new packets in
the future (for example ping and so on)

    Copyright 2004,2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include "show_dump.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define __USE_BSD
    #define __LITTLE_ENDIAN 1234
    #define __BYTE_ORDER    __LITTLE_ENDIAN

    typedef uint8_t     u_int8_t;
    typedef uint16_t    u_int16_t;
    typedef uint32_t    u_int32_t;

    #include "ip.h"
    #include "ip_icmp.h"

    #define close   closesocket
    #define sleep   Sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netdb.h>
    #include <netinet/in.h>
    #include <netinet/ip.h>
    #include <netinet/ip_icmp.h>
#endif



#define VER         "0.2"
#define IPSZ        sizeof(struct ip)
#define ICMPSZ      sizeof(struct icmp)
#define PCKSZ       IPSZ + ICMPSZ
#define TIMEOUT     3



void sndrcv_icmp(int type);
char *showtime(uint32_t hhmmss);
uint32_t bswap32(uint32_t value);
u_short in_cksum(void *data, int len);
int timeout(int sock);
u_int resolv(char *host);
void std_err(void);



int     sd;
struct  sockaddr_in peer;



int main(int argc, char *argv[]) {
    int     type = -1;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);

    fputs("\n"
        "Icmpinfo "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <host> [type]\n"
            "\n"
            "Launch this tool two times to know the exact remote time and type of stack\n"
            "\n", argv[0]);
        exit(1);
    }

    if(argc > 2) type = atoi(argv[2]);

    peer.sin_addr.s_addr = resolv(argv[1]);
    peer.sin_port        = 0;
    peer.sin_family      = AF_INET;

    printf("- Host:    %s\n", inet_ntoa(peer.sin_addr));

    sd = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP);
    if(sd < 0) std_err();

    if(type < 0) {
        printf(
            "- remote UTC time (type %d)\n"
            "---------------------------\n", ICMP_TSTAMP);
        sndrcv_icmp(ICMP_TSTAMP);       /* timestamp */

        printf(
            "- remote netmask (type %d)\n"
            "--------------------------\n", ICMP_MASKREQ);
        sndrcv_icmp(ICMP_MASKREQ);      /* netmask */
    } else {

        sndrcv_icmp(type);
    }

    close(sd);
    return(0);
}



void sndrcv_icmp(int type) {
    struct  icmp    *icmp;
    int     len,
            rerecv = 0;
    u_char  buff[256];  // PCKSZ was enough

    icmp = (struct icmp *)(buff + IPSZ);

        /* memset is used to initialize all the packet */
    memset(icmp, 0, ICMPSZ);
    icmp->icmp_type  = type;
    icmp->icmp_code  = 0;
    icmp->icmp_cksum = htons(in_cksum(icmp, ICMPSZ));

    if(sendto(sd, buff + IPSZ, ICMPSZ, 0, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();

redo:
    if(timeout(sd) < 0) {
        printf("- timeout: host has not replied\n");
        return;
    }
    len = recvfrom(sd, buff, PCKSZ, 0, NULL, NULL);
    if(len < 0) std_err();

    switch(type) {
        case ICMP_TSTAMP: {
            if(icmp->icmp_type == ICMP_TSTAMPREPLY) {
            /* different stacks */

                icmp->icmp_otime = ntohl(icmp->icmp_otime);
                icmp->icmp_rtime = ntohl(icmp->icmp_rtime);
                icmp->icmp_ttime = ntohl(icmp->icmp_ttime);

                printf("           0x%08x (otime)   0x%08x (rtime)   0x%08x (ttime)\n",
                    icmp->icmp_otime,
                    icmp->icmp_rtime,
                    icmp->icmp_ttime);

            /* WIN */
                printf("UTC time:  %s\t(stack type 1 [*BSD, *nix])\n",
                    showtime(icmp->icmp_ttime));

            /* BSD */
                printf("UTC time:  %s\t(stack type 2 [Win])\n",
                    showtime(bswap32(icmp->icmp_ttime)));

                rerecv = 0;

            } else {
                printf("Error: Bad packet received (type %d), no timestamp\n", type);
//                show_dump(buff, len, stdout);
                rerecv = 1;
            }
            } break;

        case ICMP_MASKREQ: {
            if(icmp->icmp_type == ICMP_MASKREPLY) {
                printf("Netmask:   %s\n",
                    inet_ntoa(*(struct in_addr *)&icmp->icmp_mask));
                rerecv = 0;

            } else {
                printf("Error: Bad packet received (type %d), no netmask\n", type);
//                show_dump(buff, len, stdout);
                rerecv = 1;
            }
            } break;

        default: {
            if(len >= IPSZ) {
                printf("IP header:\n");
                show_dump(buff, IPSZ, stdout);
                printf("ICMP header (type %d):\n", icmp->icmp_type);
                show_dump(buff + IPSZ, len - IPSZ, stdout);
            }
            if(type == icmp->icmp_type) {
                rerecv = 1;
            } else {
                rerecv = 0;
            }
            } break;
    }

    printf("\n");
    sleep(0);   // rarely useful to avoid subsequent network congestion
    if(rerecv) goto redo;
}



char *showtime(uint32_t hhmmss) {
    div_t   res;
    u_int   hh,
            mm,
            ss,
            ms;
    static char timestr[20];

    if(hhmmss > 0x5265bff) {
        // exist some OS (adsl routers) which return an invalid timestamp
        // this timestamp is available in milliseconds (confirmed)
        // but I'm not able to read the hours correctly
        // example of some of these hhmmss:
        // 0xffefdfd1 -> 0xffe96fb1 -> 0xffd536be
        hhmmss &= 0x7fffffff;   // makes it positive if it's negative
                                // at least for not failing the divide
    }

    res = div(hhmmss, 1000);
    ss  = res.quot;
    ms  = res.rem;

    res = div(ss, 3600);
    hh  = res.quot;
    ss  = res.rem;

    res = div(ss, 60);
    mm  = res.quot;
    ss  = res.rem;

    sprintf(timestr, "%02u:%02u:%02u %04d", hh, mm, ss, ms);
    return(timestr);
}



uint32_t bswap32(uint32_t value) {
    return((value & 0xff000000) >> 24) |
          ((value & 0x00ff0000) >>  8) |
          ((value & 0x0000ff00) <<  8) |
          ((value & 0x000000ff) << 24);
}



u_short in_cksum(void *data, int len) {
    u_int   sum    = 0;
    int     i      = len >> 1,
            endian = 1; // big endian
    u_short crc,
            *p     = (u_short *)data;

    if(*(char *)&endian) endian = 0;
    while(i--) sum += *p++;
    if(len & 1) sum += *p & (endian ? 0xff00 : 0xff);
    crc = sum = (sum >> 16) + (sum & 0xffff);
    if(sum >>= 16) crc += sum;
    if(!endian) crc = (crc >> 8) | (crc << 8);
    return(~crc);
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)(hp->h_addr);
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif

